import django
from optparse import NO_DEFAULT

from django.apps import apps
from django.core.management.commands.migrate import Command as MigrateCommand
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.recorder import MigrationRecorder
from django.db import connection
from django.conf import settings

from tenant_schemas.utils import get_public_schema_name, schema_exists
from hobo.multitenant.middleware import TenantMiddleware, TenantNotFound
from hobo.multitenant.management.commands import SyncCommon


class MigrateSchemasCommand(SyncCommon):
    help = "Updates database schema. Manages both apps with migrations and those without."

    def __init__(self, stdout=None, stderr=None, no_color=False):
        """
        Changes the option_list to use the options from the wrapped migrate command.
        """
        if django.VERSION <= (1, 10, 0):
            self.option_list += MigrateCommand.option_list
        super(Command, self).__init__(stdout, stderr, no_color)

    def add_arguments(self, parser):
        super(MigrateSchemasCommand, self).add_arguments(parser)
        command = MigrateCommand()
        command.add_arguments(parser)
        parser.set_defaults(verbosity=0)

    def handle(self, *args, **options):
        super(MigrateSchemasCommand, self).handle(*args, **options)
        self.PUBLIC_SCHEMA_NAME = get_public_schema_name()

        if self.sync_public and not self.schema_name:
            self.schema_name = self.PUBLIC_SCHEMA_NAME

        if self.sync_public:
            self.run_migrations(self.schema_name, settings.SHARED_APPS)
        if self.sync_tenant:
            if self.schema_name and self.schema_name != self.PUBLIC_SCHEMA_NAME:
                if not schema_exists(self.schema_name):
                    raise RuntimeError('Schema "{}" does not exist'.format(
                        self.schema_name))
                else:
                    self.run_migrations(self.schema_name, settings.TENANT_APPS)
            else:
                app_labels = []
                for app in apps.get_app_configs():
                    if app.name in settings.TENANT_APPS:
                        app_labels.append(app.label)
                loader = MigrationLoader(connection)
                loader.load_disk()
                recorder = MigrationRecorder(connection)
                applied_public_migrations = set(
                        [(app, migration) for app, migration
                         in recorder.applied_migrations()
                         if app in app_labels and
                            (app, migration) in loader.disk_migrations])
                for tenant in TenantMiddleware.get_tenants():
                    connection.set_schema(tenant.schema_name, include_public=False)
                    applied_migrations = self.get_applied_migrations(app_labels)
                    if options.get('fake') or options.get('migration_name') or options.get('app_label'):
                        # never skip migrations if explicit migration actions
                        # are given.
                        applied_migrations = []
                    if all([x in applied_migrations for x in applied_public_migrations]):
                        if int(self.options.get('verbosity', 1)) >= 1:
                            self._notice("=== Skipping migrations of schema %s" % tenant.schema_name)
                        continue
                    self.run_migrations(tenant.schema_name, settings.TENANT_APPS)

    def get_applied_migrations(self, app_labels):
        applied_migrations = []
        with connection.cursor() as cursor:
            cursor.execute('SELECT app, name FROM django_migrations')
            for row in cursor.fetchall():
                applied_migrations.append(row)
        applied_migrations = [x for x in applied_migrations if x[0] in app_labels]
        return applied_migrations

    def run_migrations(self, schema_name, included_apps):
        if int(self.options.get('verbosity', 1)) >= 1:
            self._notice("=== Running migrate for schema %s" % schema_name)
        connection.set_schema(schema_name, include_public=False)
        command = MigrateCommand()
        command.execute(*self.args, **self.options)
        connection.set_schema_to_public()

    def _notice(self, output):
        self.stdout.write(self.style.NOTICE(output))

Command = MigrateSchemasCommand
